#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <assert.h>
#include "kd.h"
#include "smooth1.h"


#define PRUNE_STEPS		5


#ifdef __hpux
#define MAXDOUBLE      1.7976931348623157e+308
#define HUGE MAXDOUBLE
#endif


void usage(void)
{
	fprintf(stderr,"USAGE:\n");
	fprintf(stderr,"denmax -e <fEpsilon> [-s <nSmooth>] [-d <fMinDensity>]\n");
	fprintf(stderr,"       [-t <fMaxTemperature>] [-z <fRedShift>]\n");
	fprintf(stderr,"       [-G <fGravConst>] [-O <fOmega>] [-H <fHubble>]\n");
	fprintf(stderr,"       [-m <nMinMembers>] [-r] [-nu]\n");
	fprintf(stderr,"       [-p <xyzPeriod>]\n");
	fprintf(stderr,"       [-px <xPeriod>] [-py <yPeriod>] [-pz <zPeriod>]\n");
	fprintf(stderr,"       [-c <xyzCenter>]\n");
	fprintf(stderr,"       [-cx <xCenter>] [-cy <yCenter>] [-cz <zCenter>]\n");
	fprintf(stderr,"       [-o <Output Name>]\n");
	exit(1);
	}

void main(int argc,char **argv)
{
	KD kd;
	SMX smx;
	int nBucket,nSmooth,i,j,nActive,nIttr,nMembers,bOutRay,bNoUnbind;
	char ach[80],achFile[80];
	float fPeriod[3],fCenter[3],fDensMin,fTempMax,eps,fStep,z,G,Omega0,H0;
	int sec1,usec1;
	int sec2,usec2;
	int sec3,usec3;
	int sec4,usec4;
	int sec5,usec5;
	

   	nBucket = 16;
	nSmooth = 64;
	nMembers = 4;
	bOutRay = 0;
	bNoUnbind = 0;
	eps = -1.0;
	z = 0.0;
	G = 1.0;
	Omega0 = 1.0;
	H0 = 0.0;
	fDensMin = 0.0;
	fTempMax = HUGE;
	strcpy(ach,"denmax");
	i = 1;
	for (j=0;j<3;++j) {
		fPeriod[j] = HUGE;
		fCenter[j] = 0.0;
		}
	while (i < argc) {
	    if (!strcmp(argv[i],"-s")) {
			++i;
			nSmooth = atoi(argv[i]);
			++i;
			}
		else if (!strcmp(argv[i],"-d")) {
			++i;
			fDensMin = atof(argv[i]);
			++i;
			}
		else if (!strcmp(argv[i],"-t")) {
			++i;
			fTempMax = atof(argv[i]);
			++i;
			}
		else if (!strcmp(argv[i],"-e")) {
			++i;
			eps = atof(argv[i]);
			++i;
			}
		else if (!strcmp(argv[i],"-z")) {
			++i;
			z = atof(argv[i]);
			++i;
			}
		else if (!strcmp(argv[i],"-G")) {
			++i;
			G = atof(argv[i]);
			++i;
			}
		else if (!strcmp(argv[i],"-O")) {
			++i;
			Omega0 = atof(argv[i]);
			++i;
			}
		else if (!strcmp(argv[i],"-o")) {
			++i;
			strcpy(ach,argv[i]);
			++i;
			}
		else if (!strcmp(argv[i],"-r")) {
			bOutRay = 1;
			++i;
			}
		else if (!strcmp(argv[i],"-nu")) {
			bNoUnbind = 1;
			++i;
			}
		else if (!strcmp(argv[i],"-H")) {
			++i;
			H0 = atof(argv[i]);
			++i;
			}
		else if (!strcmp(argv[i],"-m")) {
			++i;
			nMembers = atoi(argv[i]);
			++i;
			}
		else if (!strcmp(argv[i],"-p")) {
			++i;
			fPeriod[0] = atof(argv[i]);
			fPeriod[1] = atof(argv[i]);
			fPeriod[2] = atof(argv[i]);
			++i;
			}
		else if (!strcmp(argv[i],"-px")) {
			++i;
			fPeriod[0] = atof(argv[i]);
			++i;
			}
		else if (!strcmp(argv[i],"-py")) {
			++i;
			fPeriod[1] = atof(argv[i]);
			++i;
			}
		else if (!strcmp(argv[i],"-pz")) {
			++i;
		    fPeriod[2] = atof(argv[i]);
			++i;
			}
		else if (!strcmp(argv[i],"-c")) {
			++i;
			fCenter[0] = atof(argv[i]);
			fCenter[1] = atof(argv[i]);
			fCenter[2] = atof(argv[i]);
			++i;
			}
		else if (!strcmp(argv[i],"-cx")) {
			++i;
			fCenter[0] = atof(argv[i]);
			++i;
			}
		else if (!strcmp(argv[i],"-cy")) {
			++i;
			fCenter[1] = atof(argv[i]);
			++i;
			}
		else if (!strcmp(argv[i],"-cz")) {
			++i;
		    fCenter[2] = atof(argv[i]);
			++i;
			}
		else usage();
		}
	if (eps < 0.0) usage();
	kdInit(&kd,nBucket,fPeriod,fCenter);
	kdReadTipsy(kd,stdin);
	kdScatterActive(kd);
	kdBuildTree(kd);
	smInit(&smx,kd,nSmooth);
	kdTime(kd,&sec1,&usec1);
	smDensityInit(smx);
	kdTime(kd,&sec1,&usec1);
/*
	kdOutDensity(kd,"density.out");
*/
	kdTime(kd,&sec2,&usec2);
	kdInitMove(kd,fDensMin,fTempMax,eps);
	kdBuildMoveTree(kd);
	smAccDensity(smx);
	fStep = eps/4.0;
	kdMoveParticles(kd,fStep);
	kdScatterCut(kd);
	nIttr = 0;
	nActive = 1;
	while (nActive) {
		for (i=0;i<PRUNE_STEPS;++i) {
			kdBuildMoveTree(kd);
			smAccDensity(smx);
			kdMoveParticles(kd,fStep);
			}
		nActive = kdPruneInactive(kd,2*fStep);
		printf("Ittr:%d nActive:%d\n",nIttr,nActive);
		++nIttr;
		}
	kdTime(kd,&sec2,&usec2);

	kdTime(kd,&sec3,&usec3);
	kdFoF(kd,eps);
	kdTime(kd,&sec3,&usec3);
	/*
	 ** Micro-stepping phase.
	 */
	kdTime(kd,&sec5,&usec5);

	kdReactivateMove(kd);
	for (i=0;i<PRUNE_STEPS;++i) {
		kdBuildMoveTree(kd);
		smAccDensity(smx);
		kdMoveParticles(kd,fStep/10.0);
		}

	kdTime(kd,&sec5,&usec5);
	if (bOutRay) {
		strcpy(achFile,ach);
		strcat(achFile,".ray");
		kdOutVector(kd,achFile);
		}
	GroupOrder(kd);
	kdTooSmall(kd,nMembers);
	if (!bNoUnbind) {
		kdTime(kd,&sec4,&usec4);
		kdSetUniverse(kd,G,Omega0,H0,z);
		kdUnbind(kd,eps,2.0*eps);
		kdTime(kd,&sec4,&usec4);
		kdTooSmall(kd,nMembers);
		}
	strcpy(achFile,ach);
	strcat(achFile,".grp");
	kdOutGroup(kd,achFile);
	strcpy(achFile,ach);
	strcat(achFile,".gtp");
	kdWriteGroup(kd,achFile);
	smFinish(smx);
	printf("DENMAX CPU Time:\n");
	printf("   Initial Density:    %d.%06d\n",sec1,usec1);
	printf("   Moving Particles:   %d.%06d\n",sec2,usec2);
	printf("   Friends of Friends: %d.%06d\n",sec3,usec3);
	printf("   Microstepping:      %d.%06d\n",sec5,usec5);
	if (!bNoUnbind) printf("   Unbinding:          %d.%06d\n",sec4,usec4);
	kdFinish(kd);
	}
	

